"""Bring API exceptions."""


class BringException(Exception):
    """General exception occurred."""


class BringAuthException(BringException):
    """When an authentication error is encountered."""


class BringRequestException(BringException):
    """When the HTTP request fails."""


class BringParseException(BringException):
    """When parsing the response of a request fails."""


class BringEMailInvalidException(BringException):
    """When checkemail returns emailValid false ."""


class BringUserUnknownException(BringException):
    """When checkemail returns userExists false."""


class BringTranslationException(BringException):
    """When translating an article fails."""



"""Unit tests for bring-api."""

import aiohttp
from aioresponses import aioresponses
from dotenv import load_dotenv
import pytest

from bring_api.bring import Bring

load_dotenv()
UUID = "00000000-00000000-00000000-00000000"

BRING_LOGIN_RESPONSE = {
    "uuid": UUID,
    "publicUuid": UUID,
    "email": "EMAIL",
    "name": "NAME",
    "photoPath": "",
    "bringListUUID": UUID,
    "access_token": "ACCESS_TOKEN",
    "refresh_token": "REFRESH_TOKEN",
    "token_type": "Bearer",
    "expires_in": 604799,
}

BRING_USER_ACCOUNT_RESPONSE = {
    "userUuid": "{user_uuid}",
    "publicUserUuid": "{public_uuid}",
    "email": "{email}",
    "emailVerified": True,
    "name": "{user_name}",
    "photoPath": "bring/user/portrait/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxx",
    "userLocale": {"language": "de", "country": "DE"},
    "premiumConfiguration": {
        "hasPremium": False,
        "hideSponsoredProducts": False,
        "hideSponsoredTemplates": False,
        "hideSponsoredPosts": False,
        "hideSponsoredCategories": False,
        "hideOffersOnMain": False,
    },
}

BRING_USER_SETTINGS_RESPONSE = {
    "usersettings": [
        {"key": "autoPush", "value": "ON"},
        {"key": "purchaseStyle", "value": "grouped"},
        {"key": "premiumHideSponsoredCategories", "value": "OFF"},
        {"key": "premiumHideInspirationsBadge", "value": "OFF"},
        {"key": "premiumHideOffersBadge", "value": "OFF"},
        {"key": "premiumHideOffersOnMain", "value": "OFF"},
        {"key": "defaultListUUID", "value": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxx"},
        {"key": "discountActivatorOnMainEnabled", "value": "OFF"},
        {"key": "onboardClient", "value": "android"},
    ],
    "userlistsettings": [
        {
            "listUuid": UUID,
            "usersettings": [
                {
                    "key": "listSectionOrder",
                    "value": '["Früchte & Gemüse","Brot & Gebäck","Milch & Käse","Fleisch & Fisch","Zutaten & Gewürze","Fertig- & Tiefkühlprodukte","Getreideprodukte","Snacks & Süsswaren","Getränke & Tabak","Haushalt & Gesundheit","Pflege & Gesundheit","Tierbedarf","Baumarkt & Garten","Eigene Artikel"]',
                },
                {"key": "listArticleLanguage", "value": "de-DE"},
            ],
        }
    ],
}

BRING_LOAD_LISTS_RESPONSE = {
    "lists": [
        {
            "listUuid": UUID,
            "name": "Einkauf",
            "theme": "ch.publisheria.bring.theme.home",
        },
    ]
}

BRING_GET_LIST_RESPONSE = {
    "uuid": UUID,
    "status": "SHARED",
    "items": {
        "purchase": [
            {
                "uuid": "43bdd5a2-740a-4230-8b27-d0bbde886da7",
                "specification": "grün",
                "itemId": "Paprika",
            },
            {
                "uuid": "2de9d1c0-c211-4129-b6c5-c1260c3fc735",
                "specification": "gelb",
                "itemId": "Zucchetti",
            },
        ],
        "recently": [
            {
                "uuid": "5681ed79-c8e4-4c8b-95ec-112999d016c0",
                "specification": "rot",
                "itemId": "Paprika",
            },
            {
                "uuid": "01eea2cd-f433-4263-ad08-3d71317c4298",
                "specification": "",
                "itemId": "Pouletbrüstli",
            },
        ],
    },
}

BRING_GET_ALL_ITEM_DETAILS_RESPONSE = [
    {
        "uuid": "bfb5634c-d219-4d66-b68e-1388e54f0bb0",
        "itemId": "Milchreis",
        "listUuid": UUID,
        "userIconItemId": "Reis",
        "userSectionId": "Getreideprodukte",
        "assignedTo": "",
        "imageUrl": "",
    },
    {
        "uuid": "0056b23c-7fc3-44da-8c34-426f8b632220",
        "itemId": "Zitronensaft",
        "listUuid": UUID,
        "userIconItemId": "Zitrone",
        "userSectionId": "Zutaten & Gewürze",
        "assignedTo": "",
        "imageUrl": "",
    },
]

BRING_TOKEN_RESPONSE = {
    "access_token": "{access_token}",
    "refresh_token": "{refresh_token}",
    "token_type": "Bearer",
    "expires_in": 604799,
}


@pytest.fixture(name="session")
async def aiohttp_client_session():
    """Create  a client session."""
    async with aiohttp.ClientSession() as session:
        yield session


@pytest.fixture(name="bring")
async def bring_api_client(session):
    """Create Bring instance."""
    bring = Bring(session, "EMAIL", "PASSWORD")
    return bring


@pytest.fixture(name="mocked")
def aioclient_mock():
    """Mock Aiohttp client requests."""
    with aioresponses() as m:
        yield m



"""Bring api implementation."""

import asyncio
from http import HTTPStatus
import json
from json import JSONDecodeError
import logging
import os
import time
import traceback
from typing import cast

import aiohttp

from .const import (
    API_BASE_URL,
    BRING_DEFAULT_LOCALE,
    BRING_SUPPORTED_LOCALES,
    DEFAULT_HEADERS,
    LOCALES_BASE_URL,
    MAP_LANG_TO_LOCALE,
)
from .exceptions import (
    BringAuthException,
    BringEMailInvalidException,
    BringParseException,
    BringRequestException,
    BringTranslationException,
    BringUserUnknownException,
)
from .types import (
    BringAuthResponse,
    BringAuthTokenResponse,
    BringItem,
    BringItemOperation,
    BringItemsResponse,
    BringListItemDetails,
    BringListItemsDetailsResponse,
    BringListResponse,
    BringNotificationsConfigType,
    BringNotificationType,
    BringSyncCurrentUserResponse,
    BringUserListSettingEntry,
    BringUserSettingsEntry,
    BringUserSettingsResponse,
)

_LOGGER = logging.getLogger(__name__)


class Bring:
    """Unofficial Bring API interface."""

    def __init__(
        self, session: aiohttp.ClientSession, mail: str, password: str
    ) -> None:
        """Init function for Bring API."""
        self._session = session

        self.mail = mail
        self.password = password

        self.public_uuid = ""
        self.user_list_settings: dict[str, dict[str, str]] = {}
        self.user_locale = BRING_DEFAULT_LOCALE

        self.__translations: dict[str, dict[str, str]] = {}
        self.uuid = ""

        self.url = API_BASE_URL

        self.headers = DEFAULT_HEADERS.copy()

        self.loop = asyncio.get_running_loop()
        self.refresh_token = ""
        self.__expires_in: int

    @property
    def expires_in(self) -> int:
        """Refresh token expiration."""
        return max(0, self.__expires_in - int(time.time()))

    @expires_in.setter
    def expires_in(self, expires_in: int | str) -> None:
        self.__expires_in = int(time.time()) + int(expires_in)

    async def login(self) -> BringAuthResponse:
        """Try to login.

        Returns
        -------
        Response
            The server response object.

        Raises
        ------
        BringRequestException
            If the request fails.
        BringParseException
            If the parsing of the request response fails.
        BringAuthException
            If the login fails due invalid credentials.
            You should check your email and password.

        """
        user_data = {"email": self.mail, "password": self.password}

        try:
            url = f"{self.url}v2/bringauth"
            async with self._session.post(url, data=user_data) as r:
                _LOGGER.debug(
                    "Response from %s [%s]: %s",
                    url,
                    r.status,
                    await r.text()
                    if r.status != 200
                    else "",  # do not log response on success, as it contains sensible data
                )

                if r.status == HTTPStatus.UNAUTHORIZED:
                    try:
                        errmsg = await r.json()
                    except JSONDecodeError as e:
                        _LOGGER.debug(
                            "Exception: Cannot parse login request response:\n %s",
                            traceback.format_exc(),
                        )
                        raise BringParseException(
                            "Login failed due to authorization failure "
                            "but error response could not be parsed."
                        ) from e
                    _LOGGER.debug("Exception: Cannot login: %s", errmsg["message"])
                    raise BringAuthException(
                        "Login failed due to authorization failure, "
                        "please check your email and password."
                    )
                if r.status == HTTPStatus.BAD_REQUEST:
                    _LOGGER.debug("Exception: Cannot login: %s", await r.text())
                    raise BringAuthException(
                        "Login failed due to bad request, please check your email."
                    )
                r.raise_for_status()

                try:
                    data = cast(
                        BringAuthResponse,
                        {
                            key: val
                            for key, val in (await r.json()).items()
                            if key in BringAuthResponse.__annotations__
                        },
                    )
                except JSONDecodeError as e:
                    _LOGGER.debug(
                        "Exception: Cannot login:\n %s", traceback.format_exc()
                    )
                    raise BringParseException(
                        "Cannot parse login request response."
                    ) from e
        except TimeoutError as e:
            _LOGGER.debug("Exception: Cannot login:\n %s", traceback.format_exc())
            raise BringRequestException(
                "Authentication failed due to connection timeout."
            ) from e
        except aiohttp.ClientError as e:
            _LOGGER.debug("Exception: Cannot login:\n %s", traceback.format_exc())
            raise BringRequestException(
                "Authentication failed due to request exception."
            ) from e

        self.uuid = data["uuid"]
        self.public_uuid = data.get("publicUuid", "")
        self.headers["X-BRING-USER-UUID"] = self.uuid
        self.headers["Authorization"] = f'{data["token_type"]} {data["access_token"]}'
        self.refresh_token = data["refresh_token"]
        self.expires_in = data["expires_in"]

        locale = (await self.get_user_account())["userLocale"]
        self.headers["X-BRING-COUNTRY"] = locale["country"]
        self.user_locale = self.map_user_language_to_locale(locale)

        self.user_list_settings = await self.__load_user_list_settings()

        self.__translations = await self.__load_article_translations()

        return data

    async def load_lists(self) -> BringListResponse:
        """Load all shopping lists.

        Returns
        -------
        BringListResponse
            The JSON response.

        Raises
        ------
        BringRequestException
            If the request fails.
        BringParseException
            If the parsing of the request response fails.
        BringAuthException
            If the request fails due to invalid or expired authorization token.

        """
        try:
            url = f"{self.url}bringusers/{self.uuid}/lists"
            async with self._session.get(url, headers=self.headers) as r:
                _LOGGER.debug(
                    "Response from %s [%s]: %s", url, r.status, await r.text()
                )

                if r.status == HTTPStatus.UNAUTHORIZED:
                    try:
                        errmsg = await r.json()
                    except JSONDecodeError as e:
                        _LOGGER.debug(
                            "Exception: Cannot parse request response:\n %s",
                            traceback.format_exc(),
                        )
                        raise BringParseException(
                            "Loading lists failed due to authorization failure but "
                            "error response could not be parsed."
                        ) from e
                    _LOGGER.debug(
                        "Exception: Cannot get lists: %s",
                        errmsg["message"],
                    )
                    raise BringAuthException(
                        "Loading lists failed due to authorization failure, "
                        "the authorization token is invalid or expired."
                    )

                r.raise_for_status()

                try:
                    data = cast(
                        BringListResponse,
                        {
                            key: val
                            for key, val in (await r.json()).items()
                            if key in BringListResponse.__annotations__
                        },
                    )
                except JSONDecodeError as e:
                    _LOGGER.debug(
                        "Exception: Cannot get lists:\n %s", traceback.format_exc()
                    )
                    raise BringParseException(
                        "Loading lists failed during parsing of request response."
                    ) from e
                else:
                    return data
        except TimeoutError as e:
            _LOGGER.debug("Exception: Cannot get lists:\n %s", traceback.format_exc())
            raise BringRequestException(
                "Loading list failed due to connection timeout."
            ) from e
        except aiohttp.ClientError as e:
            _LOGGER.debug("Exception: Cannot get lists:\n %s", traceback.format_exc())
            raise BringRequestException(
                "Loading lists failed due to request exception."
            ) from e

    async def get_list(self, list_uuid: str) -> BringItemsResponse:
        """Get all items from a shopping list.

        Parameters
        ----------
        list_uuid : str
            A list uuid returned by load_lists()

        Returns
        -------
        BringItemsResponse
            The JSON response.

        Raises
        ------
        BringRequestException
            If the request fails.
        BringParseException
            If the parsing of the request response fails.
        BringAuthException
            If the request fails due to invalid or expired authorization token.

        """
        try:
            url = f"{self.url}v2/bringlists/{list_uuid}"
            async with self._session.get(url, headers=self.headers) as r:
                _LOGGER.debug(
                    "Response from %s [%s]: %s", url, r.status, await r.text()
                )

                if r.status == HTTPStatus.UNAUTHORIZED:
                    try:
                        errmsg = await r.json()
                    except JSONDecodeError as e:
                        _LOGGER.debug(
                            "Exception: Cannot parse request response:\n %s",
                            traceback.format_exc(),
                        )
                        raise BringParseException(
                            "Loading list items failed due to authorization failure but "
                            "error response could not be parsed."
                        ) from e
                    _LOGGER.debug(
                        "Exception: Cannot get list items: %s",
                        errmsg["message"],
                    )
                    raise BringAuthException(
                        "Loading list items failed due to authorization failure, "
                        "the authorization token is invalid or expired."
                    )

                r.raise_for_status()

                try:
                    data = await r.json()

                    for lst in data["items"].values():
                        for item in lst:
                            item["itemId"] = self.__translate(
                                item["itemId"],
                                to_locale=self.__locale(list_uuid),
                            )

                    return BringItemsResponse(
                        uuid=data["uuid"],
                        status=data["status"],
                        purchase=data["items"]["purchase"],
                        recently=data["items"]["recently"],
                    )
                except (JSONDecodeError, KeyError) as e:
                    _LOGGER.debug(
                        "Exception: Cannot get items for list %s:\n%s",
                        list_uuid,
                        traceback.format_exc(),
                    )
                    raise BringParseException(
                        "Loading list items failed during parsing of request response."
                    ) from e
        except TimeoutError as e:
            _LOGGER.debug(
                "Exception: Cannot get items for list %s:\n%s",
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                "Loading list items failed due to connection timeout."
            ) from e
        except aiohttp.ClientError as e:
            _LOGGER.debug(
                "Exception: Cannot get items for list %s:\n%s",
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                "Loading list items failed due to request exception."
            ) from e

    async def get_all_item_details(
        self, list_uuid: str
    ) -> BringListItemsDetailsResponse:
        """Get all details from a shopping list.

        Parameters
        ----------
        list_uuid : str
            A list uuid returned by load_lists()

        Returns
        -------
        BringListItemsDetailsResponse
            The JSON response. A list of item details.
            Caution: This is NOT a list of the items currently marked as 'to buy'.
            See get_list() for that.
            This contains the items that where customized by changing
            their default icon, category or uploading an image.

        Raises
        ------
        BringRequestException
            If the request fails.
        BringParseException
            If the parsing of the request response fails.
        BringAuthException
            If the request fails due to invalid or expired authorization token.

        """
        try:
            url = f"{self.url}bringlists/{list_uuid}/details"
            async with self._session.get(url, headers=self.headers) as r:
                _LOGGER.debug(
                    "Response from %s [%s]: %s", url, r.status, await r.text()
                )

                if r.status == HTTPStatus.UNAUTHORIZED:
                    try:
                        errmsg = await r.json()
                    except JSONDecodeError as e:
                        _LOGGER.debug(
                            "Exception: Cannot parse request response:\n %s",
                            traceback.format_exc(),
                        )
                        raise BringParseException(
                            "Loading list details failed due to authorization failure but "
                            "error response could not be parsed."
                        ) from e
                    _LOGGER.debug(
                        "Exception: Cannot get list details: %s",
                        errmsg["message"],
                    )
                    raise BringAuthException(
                        "Loading list details failed due to authorization failure, "
                        "the authorization token is invalid or expired."
                    )

                r.raise_for_status()

                try:
                    data = [
                        cast(
                            BringListItemDetails,
                            {
                                key: val
                                for key, val in item.items()
                                if key in BringListItemDetails.__annotations__
                            },
                        )
                        for item in await r.json()
                    ]
                    return cast(BringListItemsDetailsResponse, data)
                except JSONDecodeError as e:
                    _LOGGER.debug(
                        "Exception: Cannot get item details for list %s:\n%s",
                        list_uuid,
                        traceback.format_exc(),
                    )
                    raise BringParseException(
                        "Loading list details failed during parsing of request response."
                    ) from e
        except TimeoutError as e:
            _LOGGER.debug(
                "Exception: Cannot get item details for list %s:\n%s",
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                "Loading list details failed due to connection timeout."
            ) from e
        except aiohttp.ClientError as e:
            _LOGGER.debug(
                "Exception: Cannot get item details for list %s:\n%s",
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                "Loading list details failed due to request exception."
            ) from e

    async def save_item(
        self,
        list_uuid: str,
        item_name: str,
        specification: str = "",
        item_uuid: str = "",
    ) -> aiohttp.ClientResponse:
        """Save an item to a shopping list.

        Parameters
        ----------
        list_uuid : str
            A list uuid returned by load_lists()
        item_name : str
            The name of the item you want to save.
        specification : str, optional
            The details you want to add to the item.
        item_uuid : str, optional
            The uuid for the item to add. If a unique identifier is
            required it is recommended to generate a random uuid4.


        Returns
        -------
        Response
            The server response object.

        Raises
        ------
        BringRequestException
            If the request fails.

        """
        data = BringItem(
            itemId=item_name,
            spec=specification,
            uuid=item_uuid,
        )
        try:
            return await self.batch_update_list(list_uuid, data, BringItemOperation.ADD)
        except BringRequestException as e:
            _LOGGER.debug(
                "Exception: Cannot save item %s (%s) to list %s:\n%s",
                item_name,
                specification,
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                f"Saving item {item_name} ({specification}) to list {list_uuid} "
                "failed due to request exception."
            ) from e

    async def update_item(
        self,
        list_uuid: str,
        item_name: str,
        specification: str = "",
        item_uuid: str = "",
    ) -> aiohttp.ClientResponse:
        """Update an existing list item.

        Caution: Do not update `item_name`. Providing `item_uuid` makes it
        possible to update a specific item in case there are multiple
        items with the same name. If uuid is not specified, the newest
        item with the given `item_name` will be updated.

        Parameters
        ----------
        list_uuid : str
            A list uuid returned by load_lists()
        item_name : str
            The name of the item you want to update.
        specification : str, optional
            The details you want to update on the item.
        item_uuid : str, optional
            The uuid of the item to update.

        Returns
        -------
        Response
            The server response object.

        Raises
        ------
        BringRequestException
            If the request fails.

        """
        data = BringItem(
            itemId=item_name,
            spec=specification,
            uuid=item_uuid,
        )
        try:
            return await self.batch_update_list(list_uuid, data, BringItemOperation.ADD)
        except BringRequestException as e:
            _LOGGER.debug(
                "Exception: Cannot update item %s (%s) to list %s:\n%s",
                item_name,
                specification,
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                f"Updating item {item_name} ({specification}) in list {list_uuid} "
                "failed due to request exception."
            ) from e

    async def remove_item(
        self, list_uuid: str, item_name: str, item_uuid: str = ""
    ) -> aiohttp.ClientResponse:
        """Remove an item from a shopping list.

        Parameters
        ----------
        list_uuid : str
            A list uuid returned by load_lists()
        item_name : str
            The name of the item you want to remove.
        item_uuid : str, optional
            The uuid of the item you want to remove. The item to remove can be remove by only
            referencing its uuid and setting item_name to any nonempty string.

        Returns
        -------
        Response
            The server response object.

        Raises
        ------
        BringRequestException
            If the request fails.

        """
        data = BringItem(
            itemId=item_name,
            spec="",
            uuid=item_uuid,
        )
        try:
            return await self.batch_update_list(
                list_uuid, data, BringItemOperation.REMOVE
            )
        except BringRequestException as e:
            _LOGGER.debug(
                "Exception: Cannot delete item %s from list %s:\n%s",
                item_name,
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                f"Removing item {item_name} from list {list_uuid} "
                "failed due to request exception."
            ) from e

    async def complete_item(
        self,
        list_uuid: str,
        item_name: str,
        specification: str = "",
        item_uuid: str = "",
    ) -> aiohttp.ClientResponse:
        """Complete an item from a shopping list. This will add it to recent items.

        If it was not on the list, it will still be added to recent items.

        Parameters
        ----------
        list_uuid : str
            A list uuid returned by load_lists()
        item_name : str
            The name of the item you want to complete.
        specification : str, optional
            The details you want to update on the item.
        item_uuid : str, optional
            The uuid of the item you want to complete.

        Returns
        -------
        Response
            The server response object.

        Raises
        ------
        BringRequestException
            If the request fails.

        """
        data = BringItem(
            itemId=item_name,
            spec=specification,
            uuid=item_uuid,
        )
        try:
            return await self.batch_update_list(
                list_uuid, data, BringItemOperation.COMPLETE
            )
        except BringRequestException as e:
            _LOGGER.debug(
                "Exception: Cannot complete item %s in list %s:\n%s",
                item_name,
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                f"Completing item {item_name} from list {list_uuid} "
                "failed due to request exception."
            ) from e

    async def notify(
        self,
        list_uuid: str,
        notification_type: BringNotificationType,
        item_name: str | None = None,
    ) -> aiohttp.ClientResponse:
        """Send a push notification to all other members of a shared list.

        Parameters
        ----------
        list_uuid : str
            A list uuid returned by loadLists()
        notification_type : BringNotificationType
            The type of notification to be sent
        item_name : str, optional
            The item_name **must** be included when notication_type
            is BringNotificationType.URGENT_MESSAGE

        Returns
        -------
        Response
            The server response object.

        Raises
        ------
        BringRequestException
            If the request fails.
        BringAuthException
            If the request fails due to invalid or expired authorization token.
        TypeError
            if the notification_type parameter is invalid.
        ValueError
            If the value for item_name is invalid.

        """
        json_data = BringNotificationsConfigType(
            arguments=[],
            listNotificationType=notification_type.value,
            senderPublicUserUuid=self.public_uuid,
        )

        if not isinstance(notification_type, BringNotificationType):
            raise TypeError(
                f"notificationType {notification_type} not supported,"
                "must be of type BringNotificationType."
            )
        if notification_type is BringNotificationType.URGENT_MESSAGE:
            if not item_name or len(item_name) == 0:
                raise ValueError(
                    "notificationType is URGENT_MESSAGE but argument itemName missing."
                )

            json_data["arguments"] = [item_name]
        try:
            url = f"{self.url}v2/bringnotifications/lists/{list_uuid}"
            async with self._session.post(
                url, headers=self.headers, json=json_data
            ) as r:
                _LOGGER.debug(
                    "Response from %s [%s]: %s", url, r.status, await r.text()
                )

                if r.status == HTTPStatus.UNAUTHORIZED:
                    try:
                        errmsg = await r.json()
                    except JSONDecodeError as e:
                        _LOGGER.debug(
                            "Exception: Cannot parse request response:\n %s",
                            traceback.format_exc(),
                        )
                        raise BringParseException(
                            "Sending notification failed due to authorization failure but "
                            "error response could not be parsed."
                        ) from e
                    _LOGGER.debug(
                        "Exception: Cannot send notification: %s",
                        errmsg["message"],
                    )
                    raise BringAuthException(
                        "Sending notification failed due to authorization failure, "
                        "the authorization token is invalid or expired."
                    )

                r.raise_for_status()
                return r
        except TimeoutError as e:
            _LOGGER.debug(
                "Exception: Cannot send notification %s for list %s:\n%s",
                notification_type,
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                f"Sending notification {notification_type} for list {list_uuid}"
                "failed due to connection timeout."
            ) from e
        except aiohttp.ClientError as e:
            _LOGGER.debug(
                "Exception: Cannot send notification %s for list %s:\n%s",
                notification_type,
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                f"Sending notification {notification_type} for list {list_uuid}"
                "failed due to request exception."
            ) from e

    async def does_user_exist(self, mail: str | None = None) -> bool:
        """Check if e-mail is valid and user exists.

        Parameters
        ----------
        mail : str
            An e-mail address.

        Returns
        -------
        bool
            True if user exists.

        Raises
        ------
        BringRequestException
            If the request fails.
        BringEMailInvalidException
            If the email is invalid.
        BringUserUnknownException
            If the user is does not exist

        """
        mail = mail or self.mail

        if not mail:
            raise ValueError("Argument mail missing.")

        params = {"email": mail}

        try:
            url = f"{self.url}bringusers"
            async with self._session.get(url, headers=self.headers, params=params) as r:
                _LOGGER.debug(
                    "Response from %s [%s]: %s", url, r.status, await r.text()
                )

                if r.status == HTTPStatus.NOT_FOUND:
                    raise BringUserUnknownException(f"User {mail} does not exist.")

                r.raise_for_status()

        except TimeoutError as e:
            _LOGGER.debug(
                "Exception: Cannot get verification for %s:\n%s",
                mail,
                traceback.format_exc(),
            )
            raise BringRequestException(
                "Verifying email failed due to connection timeout."
            ) from e
        except aiohttp.ClientError as e:
            raise BringEMailInvalidException(f"E-mail {mail} is invalid.") from e

        return True

    def __load_article_translations_from_file(self, locale: str) -> dict[str, str]:
        """Read localization ressource files from disk.

        Parameters
        ----------
        locale : str
            A locale string

        Returns
        -------
            dict[str, str]:
                A translation table as a dict

        """
        dictionary_from_file: dict[str, str]

        path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "locales",
            f"articles.{locale}.json",
        )
        with open(path, encoding="UTF-8") as f:
            dictionary_from_file = json.load(f)

        return dictionary_from_file

    async def __load_article_translations(self) -> dict[str, dict[str, str]]:
        """Load all required translation dictionaries into memory.

        Raises
        ------
        BringRequestException
            If the request fails.
        BringParseException
            If the parsing of the request response fails.

        Returns
        -------
        dict
            dict of downloaded dictionaries

        """
        dictionaries: dict[str, dict[str, str]] = {}

        locales_required = list(
            dict.fromkeys(
                [
                    list_setting.get("listArticleLanguage", self.user_locale)
                    for list_setting in self.user_list_settings.values()
                ]
                + [self.user_locale]
            )
        )

        for locale in locales_required:
            if locale == BRING_DEFAULT_LOCALE or locale not in BRING_SUPPORTED_LOCALES:
                continue

            try:
                dictionaries[locale] = await self.loop.run_in_executor(
                    None, self.__load_article_translations_from_file, locale
                )
                continue
            except OSError:
                _LOGGER.warning(
                    "Locale file articles.%s.json could not be loaded from filesystem. "
                    "Will continue trying to download locale.",
                    locale,
                )

            try:
                url = f"{LOCALES_BASE_URL}articles.{locale}.json"
                async with self._session.get(url) as r:
                    _LOGGER.debug("Response from %s [%s]", url, r.status)
                    r.raise_for_status()

                    try:
                        dictionaries[locale] = await r.json()
                    except JSONDecodeError as e:
                        _LOGGER.debug(
                            "Exception: Cannot load articles.%s.json:\n%s",
                            locale,
                            traceback.format_exc(),
                        )
                        raise BringParseException(
                            f"Loading article translations for locale {locale} "
                            "failed during parsing of request response."
                        ) from e
            except TimeoutError as e:
                _LOGGER.debug(
                    "Exception: Cannot load articles.%s.json::\n%s",
                    locale,
                    traceback.format_exc(),
                )
                raise BringRequestException(
                    f"Loading article translations for locale {locale} "
                    "failed due to connection timeout."
                ) from e

            except aiohttp.ClientError as e:
                _LOGGER.debug(
                    "Exception: Cannot load articles.%s.json:\n%s",
                    locale,
                    traceback.format_exc(),
                )
                raise BringRequestException(
                    f"Loading article translations for locale {locale} "
                    "failed due to request exception."
                ) from e

        return dictionaries

    def __translate(
        self,
        item_id: str,
        *,
        to_locale: str | None = None,
        from_locale: str | None = None,
    ) -> str:
        """Translate a catalog item from or to a language.

        Parameters
        ----------
        item_id : str
            Item name.
        to_locale : str
            locale to translate to.
        from_locale : str
            locale to translate from.

        Returns
        -------
        str
            Translated Item name.

        Raises
        ------
        BringTranslationException
            If the translation fails.

        """

        locale = to_locale or from_locale

        if locale == BRING_DEFAULT_LOCALE:
            return item_id

        if not locale:
            raise ValueError("One of the arguments from_locale or to_locale required.")

        if locale not in BRING_SUPPORTED_LOCALES:
            _LOGGER.debug("Locale %s not supported by Bring.", locale)
            raise ValueError(f"Locale {locale} not supported by Bring.")
        try:
            return (
                self.__translations[locale].get(item_id, item_id)
                if to_locale
                else (
                    {value: key for key, value in self.__translations[locale].items()}
                ).get(item_id, item_id)
            )

        except Exception as e:
            _LOGGER.debug(
                "Exception: Cannot load translation dictionary:\n%s",
                traceback.format_exc(),
            )
            raise BringTranslationException(
                "Translation failed due to error loading translation dictionary."
            ) from e

    async def __load_user_list_settings(self) -> dict[str, dict[str, str]]:
        """Load user list settings into memory.

        Raises
        ------
        BringTranslationException
            If the user list settings could not be loaded.

        Returns
        -------
         dict[str, dict[str, str]]
            A dict of settings of the users lists

        """
        try:
            return {
                user_list_setting["listUuid"]: {
                    user_setting["key"]: user_setting["value"]
                    for user_setting in user_list_setting["usersettings"]
                }
                for user_list_setting in (await self.get_all_user_settings())[
                    "userlistsettings"
                ]
            }

        except Exception as e:
            _LOGGER.debug(
                "Exception: Cannot load user list settings:\n%s",
                traceback.format_exc(),
            )
            raise BringTranslationException(
                "Translation failed due to error loading user list settings."
            ) from e

    async def get_all_user_settings(self) -> BringUserSettingsResponse:
        """Load all user settings and user list settings.

        Returns
        -------
        BringUserSettingsResponse
            The JSON response.


        Raises
        ------
        BringRequestException
            If the request fails.
        BringParseException
            If the parsing of the request response fails.
        BringAuthException
            If the request fails due to invalid or expired authorization token.

        """
        try:
            url = f"{self.url}bringusersettings/{self.uuid}"
            async with self._session.get(url, headers=self.headers) as r:
                _LOGGER.debug(
                    "Response from %s [%s]: %s", url, r.status, await r.text()
                )

                if r.status == HTTPStatus.UNAUTHORIZED:
                    try:
                        errmsg = await r.json()
                    except JSONDecodeError as e:
                        _LOGGER.debug(
                            "Exception: Cannot parse request response:\n %s",
                            traceback.format_exc(),
                        )
                        raise BringParseException(
                            "Loading user settings failed due to authorization failure but "
                            "error response could not be parsed."
                        ) from e
                    _LOGGER.debug(
                        "Exception: Cannot get user settings: %s",
                        errmsg["message"],
                    )
                    raise BringAuthException(
                        "Loading user settings failed due to authorization failure, "
                        "the authorization token is invalid or expired."
                    )

                r.raise_for_status()

                try:
                    usersettings = [
                        cast(
                            BringUserSettingsEntry,
                            {
                                key: val
                                for key, val in item.items()
                                if key in BringUserSettingsEntry.__annotations__
                            },
                        )
                        for item in (await r.json())["usersettings"]
                    ]

                    userlistsettings = (await r.json())["userlistsettings"]
                    for i, listitem in enumerate(userlistsettings):
                        userlistsettings[i]["usersettings"] = [
                            cast(
                                BringUserSettingsEntry,
                                {
                                    key: val
                                    for key, val in item.items()
                                    if key in BringUserSettingsEntry.__annotations__
                                },
                            )
                            for item in listitem["usersettings"]
                        ]

                    userlistsettings = [
                        cast(
                            BringUserListSettingEntry,
                            {
                                key: val
                                for key, val in item.items()
                                if key in BringUserListSettingEntry.__annotations__
                            },
                        )
                        for item in userlistsettings
                    ]

                    data = cast(
                        BringUserSettingsResponse,
                        {
                            "usersettings": usersettings,
                            "userlistsettings": userlistsettings,
                        },
                    )

                except JSONDecodeError as e:
                    _LOGGER.debug(
                        "Exception: Cannot get user settings for uuid %s:\n%s",
                        self.uuid,
                        traceback.format_exc(),
                    )
                    raise BringParseException(
                        "Loading user settings failed during parsing of request response."
                    ) from e
                else:
                    return data
        except TimeoutError as e:
            _LOGGER.debug(
                "Exception: Cannot get user settings for uuid %s:\n%s",
                self.uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                "Loading user settings failed due to connection timeout."
            ) from e
        except aiohttp.ClientError as e:
            _LOGGER.debug(
                "Exception: Cannot get user settings for uuid %s:\n%s",
                self.uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                "Loading user settings failed due to request exception."
            ) from e

    def __locale(self, list_uuid: str) -> str:
        """Get list or user locale.

        Returns
        -------
        str
            The locale from userlistsettings or user.

        Raises
        ------
        BringTranslationException
            If list locale could not be determined from the userlistsettings or user.

        """
        if list_uuid in self.user_list_settings:
            return self.user_list_settings[list_uuid].get(
                "listArticleLanguage", self.user_locale
            )
        return self.user_locale

    def map_user_language_to_locale(self, user_locale: dict[str, str]) -> str:
        """Map user language to a supported locale.

        The userLocale returned from the user account settings is not necessarily one of the 20
        locales used by the Bring App but rather what the user has set as language on their phone
        and the country where they are located. Usually the locale for the lists is always returned
        from the bringusersettings API endpoint. One exception exists, when user onboarding happens
        through the webApp, then the locale for the automatically created initial list is not set.
        For other lists this does not happen, as it is not possible to create more lists in the
        webApp, only in the mobile apps.

        Parameters
        ----------
        user_locale : dict
            user locale as a dict containing `language` and `country`.

        Returns
        -------
        str
            The locale corresponding to the users language.

        """
        locale = f'{user_locale["language"]}-{user_locale["country"]}'
        # if locale is a valid and supported locale we can use it.
        if locale in BRING_SUPPORTED_LOCALES:
            return locale

        # if language and country are not valid locales, we use only the language part and
        # map it to a corresponding locale or the most common for that language.
        return MAP_LANG_TO_LOCALE.get(user_locale["language"], BRING_DEFAULT_LOCALE)

    async def get_user_account(self) -> BringSyncCurrentUserResponse:
        """Get current user account.

        Returns
        -------
        BringSyncCurrentUserResponse
            The JSON response.


        Raises
        ------
        BringRequestException
            If the request fails.
        BringParseException
            If the parsing of the request response fails.
        BringAuthException
            If the request fails due to invalid or expired authorization token.

        """
        try:
            url = f"{self.url}v2/bringusers/{self.uuid}"
            async with self._session.get(url, headers=self.headers) as r:
                _LOGGER.debug(
                    "Response from %s [%s]: %s", url, r.status, await r.text()
                )

                if r.status == HTTPStatus.UNAUTHORIZED:
                    try:
                        errmsg = await r.json()
                    except JSONDecodeError as e:
                        _LOGGER.debug(
                            "Exception: Cannot parse request response:\n %s",
                            traceback.format_exc(),
                        )
                        raise BringParseException(
                            "Loading current user settings failed due to authorization failure but "
                            "error response could not be parsed."
                        ) from e
                    _LOGGER.debug(
                        "Exception: Cannot get current user settings: %s",
                        errmsg["message"],
                    )
                    raise BringAuthException(
                        "Loading current user settings failed due to authorization failure, "
                        "the authorization token is invalid or expired."
                    )

                r.raise_for_status()

                try:
                    data = cast(
                        BringSyncCurrentUserResponse,
                        {
                            key: val
                            for key, val in (await r.json()).items()
                            if key in BringSyncCurrentUserResponse.__annotations__
                        },
                    )
                except JSONDecodeError as e:
                    _LOGGER.debug(
                        "Exception: Cannot get lists:\n %s", traceback.format_exc()
                    )
                    raise BringParseException(
                        "Loading lists failed during parsing of request response."
                    ) from e
                else:
                    return data
        except TimeoutError as e:
            _LOGGER.debug(
                "Exception: Cannot get current user settings:\n %s",
                traceback.format_exc(),
            )
            raise BringRequestException(
                "Loading current user settings failed due to connection timeout."
            ) from e
        except aiohttp.ClientError as e:
            _LOGGER.debug(
                "Exception: Cannot  current user settings:\n %s", traceback.format_exc()
            )
            raise BringRequestException(
                "Loading current user settings failed due to request exception."
            ) from e

    async def batch_update_list(
        self,
        list_uuid: str,
        items: BringItem | list[BringItem] | list[dict[str, str]],
        operation: BringItemOperation | None = None,
    ) -> aiohttp.ClientResponse:
        """Batch update items on a shopping list.

        Parameters
        ----------
        list_uuid : str
            The listUuid of the list to make the changes to
        items : BringItem or List of BringItem
            Item(s) to add, complete or remove from the list
        operation : BringItemOperation, optional
            The Operation (ADD, COMPLETE, REMOVE) to perform for the supplied items on the list.
            Parameter can be ommited, and the BringItem key 'operation' can be set to TO_PURCHASE,
            TO_RECENTLY or REMOVE. Defaults to BringItemOperation.ADD if operation is neither
            passed as parameter nor is set in the BringItem.

        Returns
        -------
        Response
            The server response object.

        Raises
        ------
        BringRequestException
            If the request fails.
        BringParseException
            If the parsing of the request response fails.
        BringAuthException
            If the request fails due to invalid or expired authorization token.

        """
        if operation is None:
            operation = BringItemOperation.ADD

        _base_params = {
            "accuracy": "0.0",
            "altitude": "0.0",
            "latitude": "0.0",
            "longitude": "0.0",
        }
        if isinstance(items, dict):
            items = [items]
        json_data = {
            "changes": [
                {
                    **_base_params,
                    **item,
                    "itemId": self.__translate(
                        item["itemId"],
                        from_locale=self.__locale(list_uuid),
                    ),
                    "operation": str(item.get("operation", operation)),
                }
                for item in items
            ],
            "sender": "",
        }

        try:
            url = f"{self.url}v2/bringlists/{list_uuid}/items"
            async with self._session.put(
                url, headers=self.headers, json=json_data
            ) as r:
                _LOGGER.debug(
                    "Response from %s [%s]: %s", url, r.status, await r.text()
                )

                if r.status == HTTPStatus.UNAUTHORIZED:
                    try:
                        errmsg = await r.json()
                    except JSONDecodeError as e:
                        _LOGGER.debug(
                            "Exception: Cannot parse request response:\n %s",
                            traceback.format_exc(),
                        )
                        raise BringParseException(
                            "Batch operation failed due to authorization failure but "
                            "error response could not be parsed."
                        ) from e
                    _LOGGER.debug(
                        "Exception: Cannot get execute batch operation: %s",
                        errmsg["message"],
                    )
                    raise BringAuthException(
                        "Batch operation failed due to authorization failure, "
                        "the authorization token is invalid or expired."
                    )

                r.raise_for_status()
                return r
        except TimeoutError as e:
            _LOGGER.debug(
                "Exception: Cannot execute batch operations for list %s:\n%s",
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                f"Batch operation for list {list_uuid} failed due to connection timeout."
            ) from e
        except aiohttp.ClientError as e:
            _LOGGER.debug(
                "Exception: Cannot execute batch operations for %s:\n%s",
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                f"Batch operation for list {list_uuid} failed due to request exception."
            ) from e

    async def retrieve_new_access_token(
        self, refresh_token: str | None = None
    ) -> BringAuthTokenResponse:
        """Refresh the access token.

        Parameters
        ----------
        refresh_token : str, optional
            The refresh token to use to retrieve a new access token

        Returns
        -------
        BringAuthTokenRespone
            The JSON response.

        Raises
        ------
        BringRequestException
            If the request fails.
        BringAuthException
            If the request fails due to invalid or expired refresh token.

        """
        refresh_token = refresh_token or self.refresh_token

        user_data = {"grant_type": "refresh_token", "refresh_token": refresh_token}
        try:
            url = f"{self.url}v2/bringauth/token"
            async with self._session.post(
                url, headers=self.headers, data=user_data
            ) as r:
                _LOGGER.debug(
                    "Response from %s [%s]: %s",
                    url,
                    r.status,
                    await r.text()
                    if r.status != 200
                    else "",  # do not log response on success, as it contains sensible data
                )
                if r.status == HTTPStatus.UNAUTHORIZED:
                    try:
                        errmsg = await r.json()
                    except JSONDecodeError as e:
                        _LOGGER.debug(
                            "Exception: Cannot parse token request response:\n %s",
                            traceback.format_exc(),
                        )
                        raise BringParseException(
                            "Retrieve new access token failed due to authorization failure but "
                            "error response could not be parsed."
                        ) from e
                    _LOGGER.debug(
                        "Exception: Cannot retrieve new access token: %s",
                        errmsg["message"],
                    )
                    raise BringAuthException(
                        "Retrieve new access token failed due to authorization failure, "
                        "the refresh token is invalid or expired."
                    )

                r.raise_for_status()

                try:
                    data = cast(
                        BringAuthTokenResponse,
                        {
                            key: val
                            for key, val in (await r.json()).items()
                            if key in BringAuthTokenResponse.__annotations__
                        },
                    )
                except JSONDecodeError as e:
                    _LOGGER.debug(
                        "Exception: Cannot retrieve new access token:\n %s",
                        traceback.format_exc(),
                    )
                    raise BringParseException(
                        "Cannot parse token request response."
                    ) from e
        except TimeoutError as e:
            _LOGGER.debug("Exception: Cannot login:\n %s", traceback.format_exc())
            raise BringRequestException(
                "Retrieve new access token failed due to connection timeout."
            ) from e
        except aiohttp.ClientError as e:
            _LOGGER.debug("Exception: Cannot login:\n %s", traceback.format_exc())
            raise BringRequestException(
                "Retrieve new access token failed due to request exception."
            ) from e

        self.headers["Authorization"] = f'{data["token_type"]} {data["access_token"]}'
        self.expires_in = data["expires_in"]

        return data

    async def set_list_article_language(
        self, list_uuid: str, language: str
    ) -> aiohttp.ClientResponse:
        """Set the article language for a specified list.

        Parameters
        ----------
        list_uuid : str
            The unique identifier for the list.
        language : str
            The language to set for the list articles.

        Returns
        -------
        aiohttp.ClientResponse
            The server response object.

        Raises
        ------
        ValueError
            If the specified language is not supported.
        BringRequestException
            If the request fails.
        BringAuthException
            If the request fails due to invalid or expired authorization token.

        """
        if language not in BRING_SUPPORTED_LOCALES:
            raise ValueError(f"Language {language} not supported.")

        url = f"{self.url}bringusersettings/{self.uuid}/{list_uuid}/listArticleLanguage"

        data = {"value": language}
        try:
            async with self._session.post(url, headers=self.headers, data=data) as r:
                _LOGGER.debug(
                    "Response from %s [%s]: %s", url, r.status, await r.text()
                )
                if r.status == HTTPStatus.UNAUTHORIZED:
                    raise BringAuthException(
                        "Set list article language failed due to authorization failure, "
                        "the authorization token is invalid or expired."
                    )
                r.raise_for_status()
                self.user_list_settings = await self.__load_user_list_settings()
                self.__translations = await self.__load_article_translations()
                return r
        except TimeoutError as e:
            _LOGGER.debug(
                "Exception: Cannot set article language to %s for list %s:\n%s",
                language,
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                "Set list article language failed due to connection timeout."
            ) from e
        except aiohttp.ClientError as e:
            _LOGGER.debug(
                "Exception: Cannot set article language to %s for list %s:\n%s",
                language,
                list_uuid,
                traceback.format_exc(),
            )
            raise BringRequestException(
                "Set list article language failed due to request exception."
            ) from e



"""Convert Articles.strings files to JSON locales.

To update locales extract folder catalog from assets/catalog in the android APK
and copy it to the same directory as this file. Run `python prepare_locales_catalog.py`
and this script will create the articles.{locale}.json files under bring-api/locales.
"""
import glob
import json
import os
import re

for filename in glob.glob("**/Articles.strings", root_dir="./catalog", recursive=True):
    with open(os.path.join(os.getcwd(), "catalog", filename), encoding="UTF-8") as f:
        file = f.read()
        result = dict(
            re.findall(
                "^(.*)=(.*)$",
                file.replace("\\", ""),
                flags=re.MULTILINE,
            )
        )
        locale = os.path.split(filename)[0]

        with open(
            f"bring_api{os.sep}locales{os.sep}articles.{locale[0:2].lower()}-{locale[3:5]}.json",
            "w",
            encoding="utf-8",
        ) as t:
            json.dump(result, t, ensure_ascii=False, indent=2)



"""Bring API package."""

__version__ = "0.9.0"

from .bring import Bring
from .exceptions import (
    BringAuthException,
    BringEMailInvalidException,
    BringParseException,
    BringRequestException,
    BringTranslationException,
    BringUserUnknownException,
)
from .types import (
    BringAuthResponse,
    BringAuthTokenResponse,
    BringItem,
    BringItemOperation,
    BringItemsResponse,
    BringListItemDetails,
    BringListItemsDetailsResponse,
    BringListResponse,
    BringNotificationsConfigType,
    BringNotificationType,
    BringSyncCurrentUserResponse,
    BringUserListSettingEntry,
    BringUserSettingsEntry,
    BringUserSettingsResponse,
)

__all__ = [
    "Bring",
    "BringAuthException",
    "BringAuthResponse",
    "BringAuthTokenResponse",
    "BringEMailInvalidException",
    "BringItem",
    "BringItemOperation",
    "BringItemsResponse",
    "BringListItemDetails",
    "BringListItemsDetailsResponse",
    "BringListResponse",
    "BringNotificationsConfigType",
    "BringNotificationType",
    "BringParseException",
    "BringRequestException",
    "BringSyncCurrentUserResponse",
    "BringTranslationException",
    "BringUserListSettingEntry",
    "BringUserSettingsEntry",
    "BringUserSettingsResponse",
    "BringUserUnknownException",
]



"""Bring API types."""

from enum import Enum, StrEnum
from typing import Literal, NotRequired, TypedDict


class BringList(TypedDict):
    """A list class. Represents a single list."""

    listUuid: str
    name: str
    theme: str


class BringAttribute(TypedDict):
    """An attribute class. Represents a single item attribute."""

    type: str
    content: dict[str, bool]


class BringPurchase(TypedDict):
    """A purchase class. Represents a single item."""

    uuid: str
    itemId: str
    specification: str
    attributes: list[BringAttribute]


class BringListItemDetails(TypedDict):
    """An item details class.

    Includes several details of an item in the context of a list.
    Caution: This does not have to be an item that is currently marked as 'to buy'.
    """

    uuid: str
    itemId: str
    listUuid: str
    userIconItemId: str
    userSectionId: str
    assignedTo: str
    imageUrl: str


class BringAuthResponse(TypedDict):
    """An auth response class."""

    uuid: str
    publicUuid: str
    email: str
    name: NotRequired[str]
    photoPath: NotRequired[str]
    bringListUUID: str
    access_token: str
    refresh_token: str
    token_type: str
    expires_in: int


class BringListResponse(TypedDict):
    """A list response class."""

    lists: list[BringList]


class BringItemsResponse(TypedDict):
    """An items response class."""

    uuid: str
    status: str
    purchase: list[BringPurchase]
    recently: list[BringPurchase]


class BringListItemsDetailsResponse(list[BringListItemDetails]):
    """A response class of a list of item details."""


class BringNotificationType(Enum):
    """Notification type.

    GOING_SHOPPING: "I'm going shopping! - Last chance for adjustments"
    CHANGED_LIST: "List changed - Check it out"
    SHOPPING_DONE: "Shopping done - you can relax"
    URGENT_MESSAGE: "Breaking news - Please get {itemName}!
    """

    GOING_SHOPPING = "GOING_SHOPPING"
    CHANGED_LIST = "CHANGED_LIST"
    SHOPPING_DONE = "SHOPPING_DONE"
    URGENT_MESSAGE = "URGENT_MESSAGE"


class BringNotificationsConfigType(TypedDict):
    """A notification config."""

    arguments: list[str]
    listNotificationType: str
    senderPublicUserUuid: str


class BringUserSettingsEntry(TypedDict):
    """A user settings class. Represents a single user setting."""

    key: str
    value: str


class BringUserListSettingEntry(TypedDict):
    """A user list settings class. Represents a single list setting."""

    listUuid: str
    usersettings: list[BringUserSettingsEntry]


class BringUserSettingsResponse(TypedDict):
    """A user settings response class."""

    usersettings: list[BringUserSettingsEntry]
    userlistsettings: list[BringUserListSettingEntry]


class BringSyncCurrentUserResponse(TypedDict):
    """A sync current user response class."""

    email: str
    emailVerified: bool
    name: NotRequired[str]
    photoPath: NotRequired[str]
    premiumConfiguration: dict[str, bool]
    publicUserUuid: str
    userLocale: dict[str, str]
    userUuid: str


class BringItemOperation(StrEnum):
    """Operation to be be executed on list items."""

    ADD = "TO_PURCHASE"
    COMPLETE = "TO_RECENTLY"
    REMOVE = "REMOVE"
    ATTRIBUTE_UPDATE = "ATTRIBUTE_UPDATE"


class BringItem(TypedDict):
    """A BringItem."""

    itemId: str
    spec: str
    uuid: str
    operation: NotRequired[
        BringItemOperation
        | Literal[
            "TO_PURCHASE",
            "TO_RECENTLY",
            "REMOVE",
            "ATTRIBUTE_UPDATE",
        ]
    ]
    attribute: NotRequired[BringAttribute]


class BringAuthTokenResponse(TypedDict):
    """A refresh token response class."""

    access_token: str
    refresh_token: str
    token_type: str
    expires_in: int



"""Unit tests for bring-api."""

import asyncio
import enum
from http import HTTPStatus
import time

import aiohttp
from dotenv import load_dotenv
import pytest

from bring_api.bring import Bring
from bring_api.const import BRING_SUPPORTED_LOCALES, DEFAULT_HEADERS
from bring_api.exceptions import (
    BringAuthException,
    BringEMailInvalidException,
    BringParseException,
    BringRequestException,
    BringTranslationException,
    BringUserUnknownException,
)
from bring_api.types import BringItem, BringItemOperation, BringNotificationType

from .conftest import (
    BRING_GET_ALL_ITEM_DETAILS_RESPONSE,
    BRING_GET_LIST_RESPONSE,
    BRING_LOAD_LISTS_RESPONSE,
    BRING_LOGIN_RESPONSE,
    BRING_TOKEN_RESPONSE,
    BRING_USER_ACCOUNT_RESPONSE,
    BRING_USER_SETTINGS_RESPONSE,
    UUID,
)

load_dotenv()


def mocked_translate(bring: Bring, item_id: str, *args, **kwargs) -> str:
    """Mock __translate method."""
    return item_id


class TestDoesUserExist:
    """Tests for does_user_exist method."""

    async def test_mail_invalid(self, mocked, bring):
        """Test does_user_exist for invalid e-mail."""
        mocked.get("https://api.getbring.com/rest/bringusers?email=EMAIL", status=400)
        with pytest.raises(BringEMailInvalidException):
            await bring.does_user_exist("EMAIL")

    async def test_unknown_user(self, mocked, bring):
        """Test does_user_exist for unknown user."""
        mocked.get("https://api.getbring.com/rest/bringusers?email=EMAIL", status=404)
        with pytest.raises(BringUserUnknownException):
            await bring.does_user_exist("EMAIL")

    async def test_mail_value_error(self, mocked, bring, monkeypatch):
        """Test does_user_exist for unknown user."""
        mocked.get(
            "https://api.getbring.com/rest/bringusers?email=", status=HTTPStatus.OK
        )
        monkeypatch.setattr(bring, "mail", None)

        with pytest.raises(ValueError, match="Argument mail missing."):
            await bring.does_user_exist()

    async def test_user_exist_with_parameter(self, mocked, bring):
        """Test does_user_exist for known user."""
        mocked.get(
            "https://api.getbring.com/rest/bringusers?email=EMAIL", status=HTTPStatus.OK
        )
        assert await bring.does_user_exist("EMAIL") is True

    async def test_user_exist_without_parameter(self, mocked, bring):
        """Test does_user_exist for known user."""
        mocked.get(
            "https://api.getbring.com/rest/bringusers?email=EMAIL",
            status=HTTPStatus.OK,
        )
        assert await bring.does_user_exist() is True

    @pytest.mark.parametrize(
        ("exception", "expected"),
        [
            (asyncio.TimeoutError, BringRequestException),
            (aiohttp.ClientError, BringEMailInvalidException),
        ],
    )
    async def test_request_exception(self, mocked, bring, exception, expected):
        """Test request exceptions."""

        mocked.get(
            "https://api.getbring.com/rest/bringusers?email=EMAIL",
            exception=exception,
        )

        with pytest.raises(expected):
            await bring.does_user_exist("EMAIL")


class TestLogin:
    """Tests for login method."""

    async def test_mail_invalid(self, mocked, bring):
        """Test login with invalid e-mail."""
        mocked.post(
            "https://api.getbring.com/rest/v2/bringauth",
            status=400,
        )
        expected = "Login failed due to bad request, please check your email."
        with pytest.raises(BringAuthException, match=expected):
            await bring.login()

    async def test_unauthorized(self, mocked, bring):
        """Test login with unauthorized user."""
        mocked.post(
            "https://api.getbring.com/rest/v2/bringauth",
            status=HTTPStatus.UNAUTHORIZED,
            payload={"message": ""},
        )
        expected = "Login failed due to authorization failure, please check your email and password."
        with pytest.raises(BringAuthException, match=expected):
            await bring.login()

    @pytest.mark.parametrize("status", [HTTPStatus.OK, HTTPStatus.UNAUTHORIZED])
    async def test_parse_exception(self, mocked, bring, status):
        """Test parse exceptions."""
        mocked.post(
            "https://api.getbring.com/rest/v2/bringauth",
            status=status,
            body="not json",
            content_type="application/json",
        )

        with pytest.raises(BringParseException):
            await bring.login()

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exceptions(self, mocked, bring, exception):
        """Test exceptions."""
        mocked.post("https://api.getbring.com/rest/v2/bringauth", exception=exception)
        with pytest.raises(BringRequestException):
            await bring.login()

    async def test_login(self, mocked, bring, monkeypatch):
        """Test login with valid user."""

        mocked.post(
            "https://api.getbring.com/rest/v2/bringauth",
            status=HTTPStatus.OK,
            payload=BRING_LOGIN_RESPONSE,
        )

        async def mocked_get_user_account(*args, **kwargs):
            """Mock get_user_account."""
            return {"userLocale": {"language": "de", "country": "DE"}}

        async def mocked__load_user_list_settings(*args, **kwargs):
            """Mock __load_user_list_settings."""
            return {UUID: {"listArticleLanguage": "de-DE"}}

        async def mocked__load_article_translations(*args, **kwargs):
            """Mock __load_article_translations."""
            return {}

        monkeypatch.setattr(Bring, "get_user_account", mocked_get_user_account)
        monkeypatch.setattr(
            Bring, "_Bring__load_user_list_settings", mocked__load_user_list_settings
        )
        monkeypatch.setattr(
            Bring,
            "_Bring__load_article_translations",
            mocked__load_article_translations,
        )

        data = await bring.login()
        assert data == BRING_LOGIN_RESPONSE
        assert bring.headers["Authorization"] == "Bearer ACCESS_TOKEN"
        assert bring.headers["X-BRING-COUNTRY"] == "DE"
        assert bring.uuid == UUID
        assert bring.public_uuid == UUID
        assert bring.user_locale == "de-DE"


class TestLoadLists:
    """Tests for load_lists method."""

    async def test_load_lists(self, bring, mocked, monkeypatch):
        """Test load_lists."""

        mocked.get(
            f"https://api.getbring.com/rest/bringusers/{UUID}/lists",
            status=HTTPStatus.OK,
            payload=BRING_LOAD_LISTS_RESPONSE,
        )
        monkeypatch.setattr(bring, "uuid", UUID)

        lists = await bring.load_lists()

        assert lists == BRING_LOAD_LISTS_RESPONSE

    @pytest.mark.parametrize("status", [HTTPStatus.OK, HTTPStatus.UNAUTHORIZED])
    async def test_parse_exception(self, mocked, bring, monkeypatch, status):
        """Test parse exceptions."""
        mocked.get(
            f"https://api.getbring.com/rest/bringusers/{UUID}/lists",
            status=status,
            body="not json",
            content_type="application/json",
        )
        monkeypatch.setattr(bring, "uuid", UUID)

        with pytest.raises(BringParseException):
            await bring.load_lists()

    async def test_unauthorized(self, mocked, bring, monkeypatch):
        """Test unauthorized exception."""
        mocked.get(
            f"https://api.getbring.com/rest/bringusers/{UUID}/lists",
            status=HTTPStatus.UNAUTHORIZED,
            payload={"message": ""},
        )
        monkeypatch.setattr(bring, "uuid", UUID)
        with pytest.raises(BringAuthException):
            await bring.load_lists()

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exception(self, mocked, bring, exception, monkeypatch):
        """Test request exceptions."""
        mocked.get(
            f"https://api.getbring.com/rest/bringusers/{UUID}/lists",
            exception=exception,
        )
        monkeypatch.setattr(bring, "uuid", UUID)

        with pytest.raises(BringRequestException):
            await bring.load_lists()


class TestNotifications:
    """Tests for notification method."""

    @pytest.mark.parametrize(
        ("notification_type", "item_name"),
        [
            (BringNotificationType.GOING_SHOPPING, ""),
            (BringNotificationType.CHANGED_LIST, ""),
            (BringNotificationType.SHOPPING_DONE, ""),
            (BringNotificationType.URGENT_MESSAGE, "WITH_ITEM_NAME"),
        ],
    )
    async def test_notify(
        self,
        bring,
        notification_type: BringNotificationType,
        item_name: str,
        mocked,
    ):
        """Test GOING_SHOPPING notification."""

        mocked.post(
            f"https://api.getbring.com/rest/v2/bringnotifications/lists/{UUID}",
            status=HTTPStatus.OK,
        )
        resp = await bring.notify(UUID, notification_type, item_name)
        assert resp.status == HTTPStatus.OK

    async def test_notify_urgent_message_item_name_missing(self, bring, mocked):
        """Test URGENT_MESSAGE notification."""
        mocked.post(
            f"https://api.getbring.com/rest/v2/bringnotifications/lists/{UUID}",
            status=HTTPStatus.OK,
        )
        with pytest.raises(
            ValueError,
            match="notificationType is URGENT_MESSAGE but argument itemName missing.",
        ):
            await bring.notify(UUID, BringNotificationType.URGENT_MESSAGE, "")

    async def test_notify_notification_type_raise_attribute_error(self, bring, mocked):
        """Test URGENT_MESSAGE notification."""

        with pytest.raises(
            AttributeError,
        ):
            await bring.notify(UUID, "STRING", "")

    async def test_notify_notification_type_raise_type_error(self, bring, mocked):
        """Test URGENT_MESSAGE notification."""

        class WrongEnum(enum.Enum):
            """Test Enum."""

            UNKNOWN = "UNKNOWN"

        with pytest.raises(
            TypeError,
            match="notificationType WrongEnum.UNKNOWN not supported,"
            "must be of type BringNotificationType.",
        ):
            await bring.notify(UUID, WrongEnum.UNKNOWN, "")

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exception(self, mocked, bring, exception):
        """Test request exceptions."""

        mocked.post(
            f"https://api.getbring.com/rest/v2/bringnotifications/lists/{UUID}",
            exception=exception,
        )

        with pytest.raises(BringRequestException):
            await bring.notify(UUID, BringNotificationType.GOING_SHOPPING)

    async def test_unauthorized(self, mocked, bring):
        """Test unauthorized exception."""
        mocked.post(
            f"https://api.getbring.com/rest/v2/bringnotifications/lists/{UUID}",
            status=HTTPStatus.UNAUTHORIZED,
            payload={"message": ""},
        )
        with pytest.raises(BringAuthException):
            await bring.notify(UUID, BringNotificationType.GOING_SHOPPING)

    async def test_parse_exception(self, mocked, bring):
        """Test parse exceptions."""
        mocked.post(
            f"https://api.getbring.com/rest/v2/bringnotifications/lists/{UUID}",
            status=HTTPStatus.UNAUTHORIZED,
            body="not json",
            content_type="application/json",
        )

        with pytest.raises(BringParseException):
            await bring.notify(UUID, BringNotificationType.GOING_SHOPPING)


class TestGetList:
    """Tests for get_list method."""

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exception(self, mocked, bring, exception):
        """Test request exceptions."""

        mocked.get(
            f"https://api.getbring.com/rest/v2/bringlists/{UUID}",
            exception=exception,
        )

        with pytest.raises(BringRequestException):
            await bring.get_list(UUID)

    async def test_unauthorized(self, mocked, bring):
        """Test unauthorized exception."""
        mocked.get(
            f"https://api.getbring.com/rest/v2/bringlists/{UUID}",
            status=HTTPStatus.UNAUTHORIZED,
            payload={"message": ""},
        )
        with pytest.raises(BringAuthException):
            await bring.get_list(UUID)

    @pytest.mark.parametrize("status", [HTTPStatus.OK, HTTPStatus.UNAUTHORIZED])
    async def test_parse_exception(self, mocked, bring, monkeypatch, status):
        """Test parse exceptions."""
        mocked.get(
            f"https://api.getbring.com/rest/v2/bringlists/{UUID}",
            status=status,
            body="not json",
            content_type="application/json",
        )
        monkeypatch.setattr(bring, "uuid", UUID)

        with pytest.raises(BringParseException):
            await bring.get_list(UUID)

    async def test_get_list(self, mocked, bring, monkeypatch):
        """Test get list."""
        mocked.get(
            f"https://api.getbring.com/rest/v2/bringlists/{UUID}",
            status=HTTPStatus.OK,
            payload=BRING_GET_LIST_RESPONSE,
        )

        monkeypatch.setattr(Bring, "_Bring__locale", lambda _, x: "de-DE")
        monkeypatch.setattr(Bring, "_Bring__translate", mocked_translate)
        monkeypatch.setattr(bring, "uuid", UUID)

        data = await bring.get_list(UUID)
        assert data == {
            "uuid": BRING_GET_LIST_RESPONSE["uuid"],
            "status": BRING_GET_LIST_RESPONSE["status"],
            **BRING_GET_LIST_RESPONSE["items"],
        }


class TestGetAllItemDetails:
    """Test for get_all_item_details method."""

    async def test_get_all_item_details(self, mocked, bring):
        """Test get_all_item_details."""
        mocked.get(
            f"https://api.getbring.com/rest/bringlists/{UUID}/details",
            status=HTTPStatus.OK,
            payload=BRING_GET_ALL_ITEM_DETAILS_RESPONSE,
        )

        data = await bring.get_all_item_details(UUID)
        assert data == BRING_GET_ALL_ITEM_DETAILS_RESPONSE

    async def test_list_not_found(self, mocked, bring):
        """Test get_all_item_details."""
        mocked.get(
            f"https://api.getbring.com/rest/bringlists/{UUID}/details",
            status=404,
            reason=f"List with uuid '{UUID}' not found",
        )

        with pytest.raises(BringRequestException):
            await bring.get_all_item_details(UUID)

    async def test_unauthorized(self, mocked, bring):
        """Test unauthorized exception."""
        mocked.get(
            f"https://api.getbring.com/rest/bringlists/{UUID}/details",
            status=HTTPStatus.UNAUTHORIZED,
            payload={"message": ""},
        )
        with pytest.raises(BringAuthException):
            await bring.get_all_item_details(UUID)

    @pytest.mark.parametrize("status", [HTTPStatus.OK, HTTPStatus.UNAUTHORIZED])
    async def test_parse_exception(self, mocked, bring, status):
        """Test parse exceptions."""
        mocked.get(
            f"https://api.getbring.com/rest/bringlists/{UUID}/details",
            status=status,
            body="not json",
            content_type="application/json",
        )

        with pytest.raises(BringParseException):
            await bring.get_all_item_details(UUID)

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exception(self, mocked, bring, exception):
        """Test request exceptions."""

        mocked.get(
            f"https://api.getbring.com/rest/bringlists/{UUID}/details",
            exception=exception,
        )

        with pytest.raises(BringRequestException):
            await bring.get_all_item_details(UUID)


async def mocked_batch_update_list(
    bring: Bring,
    list_uuid: str,
    items: BringItem,
    operation: BringItemOperation,
):
    """Mock batch_update_list."""
    return (list_uuid, items, operation)


class TestSaveItem:
    """Test for save_item method."""

    @pytest.mark.parametrize(
        ("item_name", "specification", "item_uuid"),
        [
            ("item name", "", None),
            ("item name", "specification", None),
            ("item name", "", UUID),
        ],
    )
    async def test_save_item(
        self, bring, monkeypatch, item_name, specification, item_uuid
    ):
        """Test save_item."""

        monkeypatch.setattr(Bring, "batch_update_list", mocked_batch_update_list)

        list_uuid, items, operation = await bring.save_item(
            UUID, item_name, specification, item_uuid
        )
        assert list_uuid == UUID
        expected = {"itemId": item_name, "spec": specification, "uuid": item_uuid}
        assert expected == items
        assert operation == BringItemOperation.ADD

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exception(self, mocked, bring, exception):
        """Test request exceptions."""

        mocked.put(
            f"https://api.getbring.com/rest/v2/bringlists/{UUID}/items",
            exception=exception,
        )

        with pytest.raises(BringRequestException) as exc:
            await bring.save_item(UUID, "item_name", "specification")
        assert (
            exc.value.args[0]
            == f"Saving item item_name (specification) to list {UUID} "
            "failed due to request exception."
        )


class TestUpdateItem:
    """Test for save_item method."""

    @pytest.mark.parametrize(
        ("item_name", "specification", "item_uuid"),
        [
            ("item name", "", None),
            ("item name", "specification", None),
            ("item name", "", UUID),
        ],
    )
    async def test_update_item(
        self, bring, monkeypatch, item_name, specification, item_uuid
    ):
        """Test save_item."""

        monkeypatch.setattr(Bring, "batch_update_list", mocked_batch_update_list)

        list_uuid, items, operation = await bring.update_item(
            UUID, item_name, specification, item_uuid
        )
        assert list_uuid == UUID
        expected = {"itemId": item_name, "spec": specification, "uuid": item_uuid}
        assert expected == items
        assert operation == BringItemOperation.ADD

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exception(self, mocked, bring, exception):
        """Test request exceptions."""

        mocked.put(
            f"https://api.getbring.com/rest/v2/bringlists/{UUID}/items",
            exception=exception,
        )

        with pytest.raises(BringRequestException) as exc:
            await bring.update_item(UUID, "item_name", "specification")
        assert (
            exc.value.args[0]
            == f"Updating item item_name (specification) in list {UUID} "
            "failed due to request exception."
        )


class TestRemoveItem:
    """Test for save_item method."""

    @pytest.mark.parametrize(
        ("item_name", "item_uuid"),
        [
            ("item name", None),
            ("item name", UUID),
        ],
    )
    async def test_remove_item(self, bring, monkeypatch, item_name, item_uuid):
        """Test save_item."""

        monkeypatch.setattr(Bring, "batch_update_list", mocked_batch_update_list)

        list_uuid, items, operation = await bring.remove_item(
            UUID, item_name, item_uuid
        )
        assert list_uuid == UUID
        expected = {"itemId": item_name, "spec": "", "uuid": item_uuid}
        assert expected == items
        assert operation == BringItemOperation.REMOVE

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exception(self, mocked, bring, exception):
        """Test request exceptions."""

        mocked.put(
            f"https://api.getbring.com/rest/v2/bringlists/{UUID}/items",
            exception=exception,
        )

        with pytest.raises(BringRequestException) as exc:
            await bring.remove_item(UUID, "item_name")
        assert (
            exc.value.args[0] == f"Removing item item_name from list {UUID} "
            "failed due to request exception."
        )


class TestCompleteItem:
    """Test for save_item method."""

    @pytest.mark.parametrize(
        ("item_name", "specification", "item_uuid"),
        [
            ("item name", "", None),
            ("item name", "specification", None),
            ("item name", "", UUID),
        ],
    )
    async def test_complete_item(
        self, bring, monkeypatch, item_name, specification, item_uuid
    ):
        """Test save_item."""

        monkeypatch.setattr(Bring, "batch_update_list", mocked_batch_update_list)

        list_uuid, items, operation = await bring.complete_item(
            UUID, item_name, specification, item_uuid
        )
        assert list_uuid == UUID
        expected = {"itemId": item_name, "spec": specification, "uuid": item_uuid}
        assert expected == items
        assert operation == BringItemOperation.COMPLETE

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exception(self, mocked, bring, exception):
        """Test request exceptions."""

        mocked.put(
            f"https://api.getbring.com/rest/v2/bringlists/{UUID}/items",
            exception=exception,
        )

        with pytest.raises(BringRequestException) as exc:
            await bring.complete_item(UUID, "item_name")
        assert (
            exc.value.args[0] == f"Completing item item_name from list {UUID} "
            "failed due to request exception."
        )


class TestArticleTranslations:
    """Test loading of article translation tables."""

    def mocked__load_article_translations_from_file(self, locale):
        """Mock and raise for fallback to ressource download."""
        raise OSError()

    def test_load_file(self, bring, mocked):
        """Test loading json from file."""

        dictionary = bring._Bring__load_article_translations_from_file("de-CH")

        assert "Pouletbrüstli" in dictionary
        assert dictionary["Pouletbrüstli"] == "Pouletbrüstli"
        assert len(dictionary) == 444

    async def test_load_from_list_article_language(self, bring, monkeypatch):
        """Test loading json from listArticleLanguage."""

        monkeypatch.setattr(
            bring, "user_list_settings", {UUID: {"listArticleLanguage": "de-DE"}}
        )

        dictionaries = await bring._Bring__load_article_translations()

        assert "de-DE" in dictionaries
        assert dictionaries["de-DE"]["Pouletbrüstli"] == "Hähnchenbrust"
        assert len(dictionaries["de-DE"]) == 444

    async def test_load_from_user_locale(self, bring, monkeypatch):
        """Test loading json from user_locale."""

        monkeypatch.setattr(bring, "user_locale", "de-DE")

        dictionaries = await bring._Bring__load_article_translations()

        assert "de-DE" in dictionaries
        assert dictionaries["de-DE"]["Pouletbrüstli"] == "Hähnchenbrust"
        assert len(dictionaries["de-DE"]) == 444

    @pytest.mark.parametrize(
        ("test_locale", "expected_locale"),
        [
            ("de-XX", "de-DE"),
            ("en-XX", "en-US"),
            ("es-XX", "es-ES"),
            ("de-DE", "de-DE"),
            ("en-GB", "en-GB"),
        ],
    )
    async def test_map_user_language_to_locale(
        self, bring, test_locale, expected_locale
    ):
        """Test mapping invalid user_locale to valid locale."""

        user_locale = {"language": test_locale[0:2], "country": test_locale[3:5]}
        locale = bring.map_user_language_to_locale(user_locale)

        assert expected_locale == locale

    def test_get_locale_from_list(self, bring, monkeypatch):
        """Test get locale from list."""

        monkeypatch.setattr(
            bring, "user_list_settings", {UUID: {"listArticleLanguage": "de-DE"}}
        )
        monkeypatch.setattr(bring, "user_locale", "es-ES")
        locale = bring._Bring__locale(UUID)

        assert locale == "de-DE"

    def test_get_locale_from_user(self, bring, monkeypatch):
        """Test get locale from user_locale."""

        monkeypatch.setattr(
            bring, "user_list_settings", {UUID: {"listArticleLanguage": "de-DE"}}
        )
        monkeypatch.setattr(bring, "user_locale", "es-ES")
        locale = bring._Bring__locale("xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxx")

        assert locale == "es-ES"

    def test_get_locale_from_list_fallback(self, bring, monkeypatch):
        """Test get locale from list and fallback to user_locale."""

        monkeypatch.setattr(bring, "user_list_settings", {UUID: {}})
        monkeypatch.setattr(bring, "user_locale", "es-ES")
        locale = bring._Bring__locale(UUID)

        assert locale == "es-ES"

    async def test_load_all_locales(self, bring, monkeypatch):
        """Test loading all locales."""

        user_list_settings = {
            k: {"listArticleLanguage": v} for k, v in enumerate(BRING_SUPPORTED_LOCALES)
        }

        monkeypatch.setattr(bring, "user_list_settings", user_list_settings)
        dictionaries = await bring._Bring__load_article_translations()

        assert len(dictionaries) == 19  # de-CH is skipped

    async def test_load_fallback_to_download(self, bring, mocked, monkeypatch):
        """Test loading json and fallback to download from web."""
        mocked.get(
            "https://web.getbring.com/locale/articles.de-DE.json",
            payload={"test": "test"},
            status=HTTPStatus.OK,
        )

        monkeypatch.setattr(bring, "user_locale", "de-DE")

        monkeypatch.setattr(
            Bring,
            "_Bring__load_article_translations_from_file",
            self.mocked__load_article_translations_from_file,
        )

        dictionaries = await bring._Bring__load_article_translations()

        assert dictionaries["de-DE"] == {"test": "test"}

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exceptions(self, bring, mocked, monkeypatch, exception):
        """Test loading json and fallback to download from web."""
        mocked.get(
            "https://web.getbring.com/locale/articles.de-DE.json", exception=exception
        )

        monkeypatch.setattr(bring, "user_locale", "de-DE")

        monkeypatch.setattr(
            Bring,
            "_Bring__load_article_translations_from_file",
            self.mocked__load_article_translations_from_file,
        )
        with pytest.raises(BringRequestException):
            await bring._Bring__load_article_translations()

    async def test_parse_exception(self, bring, mocked, monkeypatch):
        """Test loading json and fallback to download from web."""
        mocked.get(
            "https://web.getbring.com/locale/articles.de-DE.json",
            status=HTTPStatus.OK,
            body="not json",
            content_type="application/json",
        )

        monkeypatch.setattr(bring, "user_locale", "de-DE")

        monkeypatch.setattr(
            Bring,
            "_Bring__load_article_translations_from_file",
            self.mocked__load_article_translations_from_file,
        )
        with pytest.raises(BringParseException):
            await bring._Bring__load_article_translations()


class TestGetUserAccount:
    """Tests for get_user_account method."""

    async def test_get_user_account(self, bring, mocked, monkeypatch):
        """Test for get_user_account."""

        mocked.get(
            f"https://api.getbring.com/rest/v2/bringusers/{UUID}",
            payload=BRING_USER_ACCOUNT_RESPONSE,
            status=HTTPStatus.OK,
        )

        monkeypatch.setattr(bring, "uuid", UUID)
        data = await bring.get_user_account()

        assert data == BRING_USER_ACCOUNT_RESPONSE

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exception(self, mocked, bring, exception, monkeypatch):
        """Test request exceptions."""

        mocked.get(
            f"https://api.getbring.com/rest/v2/bringusers/{UUID}",
            exception=exception,
        )
        monkeypatch.setattr(bring, "uuid", UUID)

        with pytest.raises(BringRequestException):
            await bring.get_user_account()

    async def test_unauthorized(self, mocked, bring, monkeypatch):
        """Test unauthorized exception."""
        mocked.get(
            f"https://api.getbring.com/rest/v2/bringusers/{UUID}",
            status=HTTPStatus.UNAUTHORIZED,
            payload={"message": ""},
        )
        monkeypatch.setattr(bring, "uuid", UUID)
        with pytest.raises(BringAuthException):
            await bring.get_user_account()

    @pytest.mark.parametrize("status", [HTTPStatus.OK, HTTPStatus.UNAUTHORIZED])
    async def test_parse_exception(self, mocked, bring, monkeypatch, status):
        """Test parse exceptions."""
        mocked.get(
            f"https://api.getbring.com/rest/v2/bringusers/{UUID}",
            status=status,
            body="not json",
            content_type="application/json",
        )
        monkeypatch.setattr(bring, "uuid", UUID)

        with pytest.raises(BringParseException):
            await bring.get_user_account()


class TestGetAllUserSettings:
    """Tests for get_all_user_settings method."""

    async def test_get_all_user_settings(self, bring, mocked, monkeypatch):
        """Test for get_user_account."""

        mocked.get(
            f"https://api.getbring.com/rest/bringusersettings/{UUID}",
            payload=BRING_USER_SETTINGS_RESPONSE,
            status=HTTPStatus.OK,
        )

        monkeypatch.setattr(bring, "uuid", UUID)
        data = await bring.get_all_user_settings()

        assert data == BRING_USER_SETTINGS_RESPONSE

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exception(self, mocked, bring, exception, monkeypatch):
        """Test request exceptions."""

        mocked.get(
            f"https://api.getbring.com/rest/bringusersettings/{UUID}",
            exception=exception,
        )
        monkeypatch.setattr(bring, "uuid", UUID)

        with pytest.raises(BringRequestException):
            await bring.get_all_user_settings()

    async def test_unauthorized(self, mocked, bring, monkeypatch):
        """Test unauthorized exception."""
        mocked.get(
            f"https://api.getbring.com/rest/bringusersettings/{UUID}",
            status=HTTPStatus.UNAUTHORIZED,
            payload={"message": ""},
        )
        monkeypatch.setattr(bring, "uuid", UUID)
        with pytest.raises(BringAuthException):
            await bring.get_all_user_settings()

    @pytest.mark.parametrize("status", [HTTPStatus.OK, HTTPStatus.UNAUTHORIZED])
    async def test_parse_exception(self, mocked, bring, monkeypatch, status):
        """Test parse exceptions."""
        mocked.get(
            f"https://api.getbring.com/rest/bringusersettings/{UUID}",
            status=status,
            body="not json",
            content_type="application/json",
        )
        monkeypatch.setattr(bring, "uuid", UUID)

        with pytest.raises(BringParseException):
            await bring.get_all_user_settings()

    async def test_load_user_list_settings(self, bring, monkeypatch):
        """Test __load_user_list_settings."""

        async def mocked_get_all_user_settings(self):
            return BRING_USER_SETTINGS_RESPONSE

        monkeypatch.setattr(
            Bring, "get_all_user_settings", mocked_get_all_user_settings
        )
        data = await bring._Bring__load_user_list_settings()

        assert data[UUID]["listArticleLanguage"] == "de-DE"

    async def test_load_user_list_settings_exception(self, bring, monkeypatch):
        """Test __load_user_list_settings."""

        async def mocked_get_all_user_settings(self):
            raise BringRequestException

        monkeypatch.setattr(
            Bring, "get_all_user_settings", mocked_get_all_user_settings
        )

        with pytest.raises(BringTranslationException):
            await bring._Bring__load_user_list_settings()


class TestTranslate:
    """Test for __translate method."""

    def test_translate_to_locale(self, bring, monkeypatch):
        """Test __translate with to_locale."""
        monkeypatch.setattr(
            bring, "_Bring__translations", {"de-DE": {"Pouletbrüstli": "Hähnchenbrust"}}
        )

        item = bring._Bring__translate("Pouletbrüstli", to_locale="de-DE")

        assert item == "Hähnchenbrust"

    def test_translate_from_locale(self, bring, monkeypatch):
        """Test __translate with from_locale."""
        monkeypatch.setattr(
            bring, "_Bring__translations", {"de-DE": {"Pouletbrüstli": "Hähnchenbrust"}}
        )

        item = bring._Bring__translate("Hähnchenbrust", from_locale="de-DE")

        assert item == "Pouletbrüstli"

    def test_translate_value_error_no_locale(self, bring):
        """Test __translate with missing locale argument."""
        with pytest.raises(
            ValueError,
            match="One of the arguments from_locale or to_locale required.",
        ):
            bring._Bring__translate("item_name")

    def test_translate_value_error_unsupported_locale(self, bring):
        """Test __translate with unsupported locale."""
        locale = "en-ES"
        with pytest.raises(
            ValueError, match=f"Locale {locale} not supported by Bring."
        ):
            bring._Bring__translate("item_name", from_locale=locale)

    def test_translate_exception(self, bring):
        """Test __translate BringTranslationException."""
        with pytest.raises(BringTranslationException):
            bring._Bring__translate("item_name", from_locale="de-DE")


class TestBatchUpdateList:
    """Tests for batch_update_list."""

    @pytest.mark.parametrize(
        ("item", "operation"),
        [
            (BringItem(itemId="item0", spec="spec", uuid=""), BringItemOperation.ADD),
            (
                BringItem(itemId="item1", spec="spec", uuid="uuid"),
                BringItemOperation.COMPLETE,
            ),
            (
                BringItem(itemId="item2", spec="spec", uuid="uuid"),
                BringItemOperation.REMOVE,
            ),
            (
                BringItem(
                    itemId="item3",
                    spec="spec",
                    uuid="uuid",
                    operation=BringItemOperation.ADD,
                ),
                None,
            ),
            (
                BringItem(
                    itemId="item4",
                    spec="spec",
                    uuid="uuid",
                    operation=BringItemOperation.COMPLETE,
                ),
                None,
            ),
            (
                BringItem(
                    itemId="item5",
                    spec="spec",
                    uuid="uuid",
                    operation=BringItemOperation.REMOVE,
                ),
                None,
            ),
            (
                BringItem(
                    itemId="item6",
                    spec="spec",
                    uuid="uuid",
                    operation="TO_PURCHASE",
                ),
                None,
            ),
            (
                BringItem(
                    itemId="item7",
                    spec="spec",
                    uuid="uuid",
                    operation="TO_RECENTLY",
                ),
                None,
            ),
            (
                BringItem(
                    itemId="item8",
                    spec="spec",
                    uuid="uuid",
                    operation="REMOVE",
                ),
                None,
            ),
            (
                BringItem(
                    itemId="item9",
                    spec="spec",
                    uuid="uuid",
                ),
                None,
            ),
            (
                BringItem(
                    itemId="item10",
                    spec="spec",
                    uuid="uuid",
                    operation=BringItemOperation.COMPLETE,
                ),
                BringItemOperation.ADD,
            ),
            (
                BringItem(
                    itemId="item11",
                    spec="spec",
                    uuid="uuid",
                    operation=BringItemOperation.REMOVE,
                ),
                BringItemOperation.ADD,
            ),
            (
                BringItem(itemId="item12", spec="", uuid=""),
                BringItemOperation.ADD,
            ),
            (
                BringItem(itemId="item13", spec="", uuid="uuid"),
                BringItemOperation.ADD,
            ),
            (
                BringItem(itemId="item14", spec="spec", uuid=""),
                BringItemOperation.ADD,
            ),
            (
                BringItem(itemId="item15", spec="spec", uuid="uuid"),
                BringItemOperation.ADD,
            ),
            (
                {"itemId": "item16", "spec": "spec", "uuid": "uuid"},
                BringItemOperation.ADD,
            ),
        ],
    )
    async def test_batch_update_list_single_item(
        self, bring, mocked, monkeypatch, item, operation
    ):
        """Test batch_update_list."""
        expected = {
            "changes": [
                {
                    "accuracy": "0.0",
                    "altitude": "0.0",
                    "latitude": "0.0",
                    "longitude": "0.0",
                    "itemId": item["itemId"],
                    "spec": item["spec"],
                    "uuid": item["uuid"],
                    "operation": str(
                        item.get("operation", operation or BringItemOperation.ADD)
                    ),
                }
            ],
            "sender": "",
        }
        mocked.put(
            url := f"https://api.getbring.com/rest/v2/bringlists/{UUID}/items",
            status=HTTPStatus.OK,
        )
        monkeypatch.setattr(Bring, "_Bring__locale", lambda _, x: "de-DE")
        monkeypatch.setattr(Bring, "_Bring__translate", mocked_translate)

        if operation:
            r = await bring.batch_update_list(UUID, item, operation)
        else:
            r = await bring.batch_update_list(UUID, item)
        assert r.status == HTTPStatus.OK
        mocked.assert_called_with(
            url,
            method="PUT",
            headers=DEFAULT_HEADERS,
            data=None,
            json=expected,
        )

    async def test_batch_update_list_multiple_items(self, bring, mocked, monkeypatch):
        """Test batch_update_list."""
        test_items = [
            BringItem(
                itemId="item1",
                spec="spec1",
                uuid="uuid1",
                operation=BringItemOperation.ADD,
            ),
            BringItem(
                itemId="item2",
                spec="spec2",
                uuid="uuid2",
                operation=BringItemOperation.COMPLETE,
            ),
            BringItem(
                itemId="item3",
                spec="spec3",
                uuid="uuid3",
                operation=BringItemOperation.REMOVE,
            ),
        ]

        expected = {
            "changes": [
                {
                    "accuracy": "0.0",
                    "altitude": "0.0",
                    "latitude": "0.0",
                    "longitude": "0.0",
                    "itemId": "item1",
                    "spec": "spec1",
                    "uuid": "uuid1",
                    "operation": "TO_PURCHASE",
                },
                {
                    "accuracy": "0.0",
                    "altitude": "0.0",
                    "latitude": "0.0",
                    "longitude": "0.0",
                    "itemId": "item2",
                    "spec": "spec2",
                    "uuid": "uuid2",
                    "operation": "TO_RECENTLY",
                },
                {
                    "accuracy": "0.0",
                    "altitude": "0.0",
                    "latitude": "0.0",
                    "longitude": "0.0",
                    "itemId": "item3",
                    "spec": "spec3",
                    "uuid": "uuid3",
                    "operation": "REMOVE",
                },
            ],
            "sender": "",
        }
        mocked.put(
            url := f"https://api.getbring.com/rest/v2/bringlists/{UUID}/items",
            status=HTTPStatus.OK,
        )
        monkeypatch.setattr(Bring, "_Bring__locale", lambda _, x: "de-DE")
        monkeypatch.setattr(Bring, "_Bring__translate", mocked_translate)

        r = await bring.batch_update_list(UUID, test_items)

        assert r.status == HTTPStatus.OK
        mocked.assert_called_with(
            url,
            method="PUT",
            headers=DEFAULT_HEADERS,
            data=None,
            json=expected,
        )

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exception(self, mocked, bring, exception):
        """Test request exceptions."""

        mocked.put(
            f"https://api.getbring.com/rest/v2/bringlists/{UUID}/items",
            exception=exception,
        )

        with pytest.raises(BringRequestException):
            await bring.batch_update_list(
                UUID, BringItem(itemId="item_name", spec="spec", uuid=UUID)
            )

    async def test_unauthorized(self, mocked, bring, monkeypatch):
        """Test unauthorized exception."""
        mocked.put(
            f"https://api.getbring.com/rest/v2/bringlists/{UUID}/items",
            status=HTTPStatus.UNAUTHORIZED,
            payload={"message": ""},
        )

        with pytest.raises(BringAuthException):
            await bring.batch_update_list(
                UUID, BringItem(itemId="item_name", spec="spec", uuid=UUID)
            )

    async def test_parse_exception(self, mocked, bring, monkeypatch):
        """Test parse exceptions."""
        mocked.put(
            f"https://api.getbring.com/rest/v2/bringlists/{UUID}/items",
            status=HTTPStatus.UNAUTHORIZED,
            body="not json",
            content_type="application/json",
        )

        with pytest.raises(BringParseException):
            await bring.batch_update_list(
                UUID, BringItem(itemId="item_name", spec="spec", uuid=UUID)
            )


class TestRetrieveNewAccessToken:
    """Test for retrieve_new_access_token method."""

    async def test_retrieve_new_access_token(self, mocked, bring, monkeypatch):
        """Test retrieve_new_access_token."""
        mocked.post(
            "https://api.getbring.com/rest/v2/bringauth/token",
            status=HTTPStatus.OK,
            payload=BRING_TOKEN_RESPONSE,
        )
        monkeypatch.setattr(bring, "refresh_token", "test_refresh_token")
        monkeypatch.setattr(time, "time", lambda: 0)
        data = await bring.retrieve_new_access_token()

        assert data == BRING_TOKEN_RESPONSE
        assert bring.headers["Authorization"] == "Bearer {access_token}"
        assert bring.expires_in == BRING_TOKEN_RESPONSE["expires_in"]

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exception(self, mocked, bring, exception):
        """Test request exceptions."""

        mocked.post(
            "https://api.getbring.com/rest/v2/bringauth/token",
            exception=exception,
        )

        with pytest.raises(BringRequestException):
            await bring.retrieve_new_access_token()

    async def test_auth_exception(self, mocked, bring):
        """Test request exceptions."""

        mocked.post(
            "https://api.getbring.com/rest/v2/bringauth/token",
            status=HTTPStatus.UNAUTHORIZED,
            payload={"message": ""},
        )

        with pytest.raises(BringAuthException):
            await bring.retrieve_new_access_token()

    @pytest.mark.parametrize("status", [HTTPStatus.OK, HTTPStatus.UNAUTHORIZED])
    async def test_parse_exception(self, mocked, bring, status):
        """Test request exceptions."""

        mocked.post(
            "https://api.getbring.com/rest/v2/bringauth/token",
            status=status,
            body="not json",
            content_type="application/json",
        )

        with pytest.raises(BringParseException):
            await bring.retrieve_new_access_token()


class TestSetListArticleLanguage:
    """Tests for set_list_article_language method."""

    async def test_set_list_article_language(self, mocked, bring, monkeypatch):
        """Test set list article language."""
        mocked.post(
            f"https://api.getbring.com/rest/bringusersettings/{UUID}/{UUID}/listArticleLanguage",
            status=HTTPStatus.OK,
        )

        monkeypatch.setattr(bring, "uuid", UUID)

        async def mocked__load_user_list_settings(*args, **kwargs):
            """Mock __load_user_list_settings."""
            return {UUID: {"listArticleLanguage": "de-DE"}}

        monkeypatch.setattr(
            Bring, "_Bring__load_user_list_settings", mocked__load_user_list_settings
        )

        resp = await bring.set_list_article_language(UUID, "de-DE")
        assert resp.status == HTTPStatus.OK

    @pytest.mark.parametrize(
        "exception",
        [
            asyncio.TimeoutError,
            aiohttp.ClientError,
        ],
    )
    async def test_request_exception(self, mocked, bring, monkeypatch, exception):
        """Test request exceptions."""

        mocked.post(
            f"https://api.getbring.com/rest/bringusersettings/{UUID}/{UUID}/listArticleLanguage",
            exception=exception,
        )

        monkeypatch.setattr(bring, "uuid", UUID)

        with pytest.raises(BringRequestException):
            await bring.set_list_article_language(UUID, "de-DE")

    async def test_unauthorized(self, mocked, bring, monkeypatch):
        """Test unauthorized exception."""
        mocked.post(
            f"https://api.getbring.com/rest/bringusersettings/{UUID}/{UUID}/listArticleLanguage",
            status=HTTPStatus.UNAUTHORIZED,
        )
        monkeypatch.setattr(bring, "uuid", UUID)
        with pytest.raises(BringAuthException):
            await bring.set_list_article_language(UUID, "de-DE")

    async def test_value_error(self, bring):
        """Test ValueError exception."""

        with pytest.raises(ValueError):
            await bring.set_list_article_language(UUID, "es-CO")



"""Constants for bring-api."""
from typing import Final

API_BASE_URL: Final = "https://api.getbring.com/rest/"
DEFAULT_HEADERS: Final = {
    "Authorization": "Bearer",
    "X-BRING-API-KEY": "cof4Nc6D8saplXjE3h3HXqHH8m7VU2i1Gs0g85Sp",
    "X-BRING-CLIENT": "android",
    "X-BRING-APPLICATION": "bring",
    "X-BRING-COUNTRY": "DE",
    "X-BRING-USER-UUID": "",
}

LOCALES_BASE_URL: Final = "https://web.getbring.com/locale/"
BRING_SUPPORTED_LOCALES: Final = [
    "de-AT",
    "de-CH",
    "de-DE",
    "en-AU",
    "en-CA",
    "en-GB",
    "en-US",
    "es-ES",
    "fr-CH",
    "fr-FR",
    "hu-HU",
    "it-CH",
    "it-IT",
    "nb-NO",
    "nl-NL",
    "pl-PL",
    "pt-BR",
    "ru-RU",
    "sv-SE",
    "tr-TR",
]

MAP_LANG_TO_LOCALE = {
    "de": "de-DE",
    "en": "en-US",
    "es": "es-ES",
    "fr": "fr-FR",
    "hu": "hu-HU",
    "it": "it-IT",
    "nb": "nb-NO",
    "nl": "nl-NL",
    "pl": "pl-PL",
    "pt": "pt-BR",
    "ru": "ru-RU",
    "sv": "sv-SE",
    "tr": "tr-TR",
}

BRING_DEFAULT_LOCALE: Final = "de-CH"



"""Test script for Bring API."""

import asyncio
import logging
import os
import sys
import uuid

import aiohttp
from dotenv import load_dotenv

from bring_api.bring import Bring
from bring_api.exceptions import BringEMailInvalidException, BringUserUnknownException
from bring_api.types import BringItemOperation, BringList, BringNotificationType

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

load_dotenv()


async def test_add_complete_remove(bring: Bring, lst: BringList):
    """Test add-complete-remove for an item."""

    # Save an item with specifications to a certain shopping list
    await bring.save_item(lst["listUuid"], "Äpfel", "low fat")

    # Get all the pending items of a list
    items = await bring.get_list(lst["listUuid"])
    logging.info("List purchase items: %s", items["purchase"])

    # Check of an item
    await bring.complete_item(lst["listUuid"], items["purchase"][0]["itemId"])

    # Get all the recent items of a list
    items = await bring.get_list(lst["listUuid"])
    logging.info("List recently items: %s", items["recently"])

    # Remove an item from a list
    await bring.remove_item(lst["listUuid"], "Äpfel")

    # Get all the items of a list
    items = await bring.get_list(lst["listUuid"])
    logging.info("List all items: %s / %s", items["purchase"], items["recently"])


async def test_push_notifications(bring: Bring, lst: BringList):
    """Test sending push notifications."""

    # Send a going shopping notification
    await bring.notify(lst["listUuid"], BringNotificationType.GOING_SHOPPING)

    # Send a urgent message with argument item name
    await bring.notify(
        lst["listUuid"], BringNotificationType.URGENT_MESSAGE, "Pouletbrüstli"
    )


async def test_does_user_exist(bring: Bring):
    """Test does_user_exist."""

    rnd = str(uuid.uuid4())

    # Test invalid e-mail
    try:
        await bring.does_user_exist(f"{rnd}@gmail")
    except BringEMailInvalidException:
        logging.info("e-mail %s@gmail asserted as invalid.", rnd)

    # Test unknown user by generating random uuid
    try:
        await bring.does_user_exist(f"{rnd}@gmail.com")
    except BringUserUnknownException:
        logging.info("e-mail %s@gmail.com asserted as user unknown.", rnd)

    # Test for known existing user
    if await bring.does_user_exist():
        logging.info("e-mail %s asserted as valid and user exists", bring.mail)


async def test_translation(bring: Bring, lst: BringList):
    """Test article translations."""
    # Replace test list locale to get predictable results and
    # read back items with different locale asure catalog items where added correctly.
    locale_to = "de-DE"
    locale_from = "de-CH"

    locale_org = bring.user_list_settings[lst["listUuid"]].get(
        "listArticleLanguage", bring.user_locale
    )

    test_items = {
        "Pouletbrüstli": "Hähnchenbrust",
        "Glacé": "Eis",
        "Zucchetti": "Zucchini",
        "Gipfeli": "Croissant",
        "Pelati": "Dosentomaten",
        "Fischstäbli": "Fischstäbchen",
        "Guetzli": "Plätzchen",
        "Paprika": "Paprika",
    }
    for k, v in test_items.items():
        # Save an item an item to
        bring.user_list_settings[lst["listUuid"]]["listArticleLanguage"] = locale_to
        await bring.save_item(lst["listUuid"], v)

        # Get all the pending items of a list
        bring.user_list_settings[lst["listUuid"]]["listArticleLanguage"] = locale_from
        items = await bring.get_list(lst["listUuid"])
        item = next(ii["itemId"] for ii in items["purchase"] if ii["itemId"] == k)
        assert item == k
        logging.info("Item: %s, translation: %s", v, item)

        await bring.remove_item(lst["listUuid"], k)

    # reset locale to original value for other tests
    bring.user_list_settings[lst["listUuid"]]["listArticleLanguage"] = locale_org


async def test_batch_list_operations(bring: Bring, lst: BringList):
    """Test batch list operations."""

    # Add items with same name but different specifications
    add_items = [
        {
            "itemId": "Cilantro",
            "spec": "100g, dried",
            "uuid": str(uuid.uuid4()),
        },
        {
            "itemId": "Cilantro",
            "spec": "fresh",
            "uuid": str(uuid.uuid4()),
        },
    ]

    await bring.batch_update_list(lst["listUuid"], add_items, BringItemOperation.ADD)
    logging.info("Add %s items to list %s", len(add_items), os.environ["LIST"])

    # Get all the pending items of a list
    items = await bring.get_list(lst["listUuid"])
    logging.info("List purchase items: %s", items["purchase"])

    # Complete items on the list
    await bring.batch_update_list(
        lst["listUuid"], add_items, BringItemOperation.COMPLETE
    )
    logging.info("Complete %s items on the list %s", len(add_items), os.environ["LIST"])

    # Get all the recent items of a list
    items = await bring.get_list(lst["listUuid"])
    logging.info("List recently items: %s", items["recently"])

    # Remove items on the list
    await bring.batch_update_list(lst["listUuid"], add_items, BringItemOperation.REMOVE)
    logging.info("Remove items from the list %s: %s", os.environ["LIST"], add_items)

    # Get all the items of a list
    items = await bring.get_list(lst["listUuid"])
    logging.info("List all items: %s / %s", items["purchase"], items["recently"])

    # Batch update list with add, complete and remove operations
    item1_uuid = str(uuid.uuid4())
    item2_uuid = str(uuid.uuid4())
    add_complete_delete_items = [
        {
            "itemId": "Hähnchen",
            "spec": "gegrillt",
            "uuid": item1_uuid,
            "operation": "TO_PURCHASE",
        },
        {
            "itemId": "Hähnchenbrust",
            "spec": "gebraten",
            "uuid": item2_uuid,
            "operation": "TO_PURCHASE",
        },
        {
            "itemId": "Hähnchen",
            "spec": "",
            "uuid": item1_uuid,
            "operation": "TO_RECENTLY",
        },
        {
            "itemId": "Hähnchenbrust",
            "spec": "",
            "uuid": item2_uuid,
            "operation": "TO_RECENTLY",
        },
        {
            "itemId": "Hähnchen",
            "spec": "",
            "uuid": item1_uuid,
            "operation": "REMOVE",
        },
        {
            "itemId": "Hähnchenbrust",
            "spec": "",
            "uuid": item2_uuid,
            "operation": "REMOVE",
        },
    ]
    await bring.batch_update_list(lst["listUuid"], add_complete_delete_items)
    logging.info(
        "Submit batch update with ADD, COMPLETE, REMOVE operations to list %s",
        os.environ["LIST"],
    )


async def main():
    """Test Bring API."""
    async with aiohttp.ClientSession() as session:
        # Create Bring instance with email and password
        bring = Bring(session, os.environ["EMAIL"], os.environ["PASSWORD"])

        # run before login
        await test_does_user_exist(bring)

        # Login
        await bring.login()

        # Get information about all available shopping list and select one to test with
        lists = (await bring.load_lists())["lists"]
        lst = next(lst for lst in lists if lst["name"] == os.environ["LIST"])
        logging.info("Selected list: %s (%s)", lst["name"], lst["listUuid"])

        await test_add_complete_remove(bring, lst)

        await test_push_notifications(bring, lst)

        await test_translation(bring, lst)

        await test_batch_list_operations(bring, lst)

        await bring.set_list_article_language(lst["listUuid"], "es-ES")
        await bring.get_list(lst["listUuid"])
        await bring.set_list_article_language(lst["listUuid"], "de-DE")


asyncio.run(main())
